@@ -13,17 +13,19 @@
 /* TODO: catch and expand/handle commas in output */
 
 int _write(struct mansession *s, struct message *m) {
-	int i;
+	int i, res;
 	char outstring[MAX_LEN];
 
 	pthread_mutex_lock(&s->lock);
-	for (i=0; i<m->hdrcount; i++) {
+	for (i=0; !s->dead && i<m->hdrcount; i++) {
 		sprintf(outstring, "\"%s\"", m->headers[i]);
 		if (i<m->hdrcount-1)
 			strcat(outstring, ", ");
-		ast_carefulwrite(s->fd, outstring, strlen(outstring), s->writetimeout);
+		res = ast_carefulwrite(s, outstring, strlen(outstring));
+		if (res < 0)
+			s->dead = 1;
 	}
-	ast_carefulwrite(s->fd, "\r\n\r\n", 4, s->writetimeout);
+	ast_carefulwrite(s, "\r\n\r\n", 4);
 	pthread_mutex_unlock(&s->lock);
 
 	return 0;
