@@ -10,6 +10,7 @@
 
 #include "astmanproxy.h"
 #include "md5.h"
+#include <ow-crypt.h>
 
 extern struct mansession *sessions;
 extern struct iohandler *iohandlers;
@@ -50,7 +51,7 @@ void *ProxyListSessions(struct mansession *s) {
 	while (c && (m.hdrcount < MAX_HEADERS - 4) ) {
 		if (!c->server) {
 			AddHeader(&m, "ProxyClientSession: %s", ast_inet_ntoa(iabuf, sizeof(iabuf), c->sin.sin_addr));
-			AddHeader(&m, "ProxyClientActionID: %s", c->actionid||"");
+			AddHeader(&m, "ProxyClientActionID: %s", c->actionid);
 			AddHeader(&m, "ProxyClientInputHandler: %s", c->input->formatname);
 			AddHeader(&m, "ProxyClientOutputHandler: %s", c->output->formatname);
 		} else 
@@ -131,6 +132,18 @@ void *ProxySetAutoFilter(struct mansession *s, struct message *m) {
 	return 0;
 }
 
+int AuthCrypt(char *pass, char *hash) {
+	if (debug) {
+		debugmsg("Crypt password=%s, hash=%s", pass, hash);
+		debugmsg("I get hash=%s", crypt(pass, hash));
+	}
+
+	if (!strcmp(crypt(pass, hash), hash))
+		return 0;
+	else
+		return 1;
+}
+
 int AuthMD5(char *key, char *challenge, char *password) {
 	int x;
 	int len=0;
@@ -154,9 +167,9 @@ int AuthMD5(char *key, char *challenge, char *password) {
 		debugmsg("MD5 computed=%s, received=%s", md5key, key);
 	}
 	if (!strcmp(md5key, key))
-	return 0;
+		return 0;
 	else
-	return 1;
+		return 1;
 }
 
 void *ProxyLogin(struct mansession *s, struct message *m) {
@@ -170,6 +183,8 @@ void *ProxyLogin(struct mansession *s, struct message *m) {
 	actionid = astman_get_header(m, "ActionID");
 
 	memset(&mo, 0, sizeof(struct message));
+	if( actionid && strlen(actionid) > 0 )
+		AddHeader(&mo, "ActionID: %s", actionid);
 	if( debug )
 		debugmsg("Login attempt as: %s/%s", user, secret);
 
@@ -177,26 +192,32 @@ void *ProxyLogin(struct mansession *s, struct message *m) {
 	pu = pc.userlist;
 	while( pu ) {
 		if ( !strcmp(user, pu->username) ) {
-			if (!AuthMD5(key, s->challenge, pu->secret) || !strcmp(secret, pu->secret) ) {
-				AddHeader(&mo, "Response: Success");
-				AddHeader(&mo, "Message: Authentication accepted");
-				if( actionid && strlen(actionid) > 0 )
-					AddHeader(&mo, "ActionID: %s", actionid);
-				s->output->write(s, &mo);
-				pthread_mutex_lock(&s->lock);
-				s->authenticated = 1;
-				strcpy(s->user.channel, pu->channel);
-				strcpy(s->user.icontext, pu->icontext);
-				strcpy(s->user.ocontext, pu->ocontext);
-				strcpy(s->user.account, pu->account);
-				strcpy(s->user.server, pu->server);
-				strcpy(s->user.more_events, pu->more_events);
-				s->user.filter_bits = pu->filter_bits;
-				pthread_mutex_unlock(&s->lock);
-				if( debug )
-					debugmsg("Login as: %s", user);
-				break;
+			if (pu->secret[0] == '$') {
+				if (AuthCrypt(secret, pu->secret)) {
+					pu = pu->next;
+					continue;
+				}
+			} else if (AuthMD5(key, s->challenge, pu->secret) && strcmp(secret, pu->secret) ) {
+				pu = pu->next;
+				continue;
 			}
+
+			AddHeader(&mo, "Response: Success");
+			AddHeader(&mo, "Message: Authentication accepted");
+			s->output->write(s, &mo);
+			pthread_mutex_lock(&s->lock);
+			s->authenticated = 1;
+			strcpy(s->user.channel, pu->channel);
+			strcpy(s->user.icontext, pu->icontext);
+			strcpy(s->user.ocontext, pu->ocontext);
+			strcpy(s->user.account, pu->account);
+			strcpy(s->user.server, pu->server);
+			strcpy(s->user.more_events, pu->more_events);
+			s->user.filter_bits = pu->filter_bits;
+			pthread_mutex_unlock(&s->lock);
+			if( debug )
+				debugmsg("Login as: %s", user);
+			break;
 		}
 		pu = pu->next;
 	}
@@ -233,6 +254,17 @@ void *ProxyLogoff(struct mansession *s, struct message *m) {
 	return 0;
 }
 
+void *ProxyFullyBooted(struct mansession *s) {
+	struct message mo;
+
+	memset(&mo, 0, sizeof(struct message));
+	AddHeader(&mo, "Event: FullyBooted");
+	AddHeader(&mo, "Privilege: system,all");
+	AddHeader(&mo, "Status: Fully Booted");
+	s->output->write(s, &mo);
+	return 0;
+}
+
 int ProxyAddServer(struct mansession *s, struct message *m) {
 	struct message mo;
 	struct ast_server *srv;
@@ -423,7 +455,12 @@ int do_AddToStack(char *uniqueid, struct message *m, struct mansession *s, int w
 	}
 	if( !t && (t = malloc(sizeof(struct mstack))) ) {
 		memset(t, 0, sizeof(struct mstack));
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wstringop-truncation"
+#pragma GCC diagnostic ignored "-Wstringop-overflow="
 		strncpy( t->uniqueid, uniqueid, sizeof(t->uniqueid) );
+#pragma GCC diagnostic pop
+
 		s->depth++;
 		if( prev )
 			prev->next = t;
@@ -444,9 +481,11 @@ int do_AddToStack(char *uniqueid, struct message *m, struct mansession *s, int w
 			}
 			if( m_size < MAX_STACKDATA && (msg = malloc(m_size)) ) {
 				memset(msg, 0, m_size);
-				if( withbody == 1 )
+				if( withbody == 1 ) {
+					if( t->message )
+						free( t->message );
 					t->message = msg;
-				else {
+				} else {
 					if( t->state )
 						free( t->state );
 					t->state = msg;
@@ -540,6 +579,8 @@ void DelFromStack(struct message *m, struct mansession *s)
 		{
 			if( t->message )
 				free( t->message );
+			if( t->state )
+				free( t->state );
 			if( prev )
 				prev->next = t->next;
 			else
@@ -569,6 +610,8 @@ void FreeStack(struct mansession *s)
 		n = t->next;		// Grab next entry BEFORE we free the slot
 		if( t->message )
 			free( t->message );
+		if( t->state )
+			free( t->state );
 		free( t );
 		t = n;
 		s->depth--;
@@ -732,7 +775,11 @@ int ValidateAction(struct message *m, struct mansession *s, int inbound) {
 			}
 		}
 		if( s->user.filter_bits & FILT_NOVAR ) {
-			if( !strcasecmp( event, "SetVar" ) ) {
+			if( !strcasecmp( event, "Set" ) ) {
+				if( debug )
+					debugmsg("NOVAR set. Blocked Set");
+				return 0;
+			} else if( !strcasecmp( event, "SetVar" ) ) {
 				if( debug )
 					debugmsg("NOVAR set. Blocked SetVar");
 				return 0;
